#
#  MyDocument.rb
#  prEdit
#
#  Created by Gregory Clarke on 6/23/10.
#  Copyright (c) 2010 Intelligent Assistance, Inc. All rights reserved.
#

require 'osx/cocoa'
require 'PlayerController'
require 'predit_algorithm.bundle'

# String constants for the toolbar
TOOLBAR_IDENTIFIER = 'My Window Toolbar Identifier'
TOOLBAR_FIRST_IDENTIFIER = 'First Toolbar Item Identifier'
TOOLBAR_SECOND_IDENTIFIER = 'Second Toolbar Item Identifier'
# Pasteboards for drag-and-drop table views
SelectsPasteboard = OSX::NSString.stringWithString "SelectsPasteboard"
StoryPasteboard = OSX::NSString.stringWithString "StoryPasteboard"
# CA Transitions
KCATransitionPush = "push"
KCATransitionFromLeft = "fromLeft"
KCATransitionFromRight = "fromRight"
KCATransitionFromTop = "fromTop"

class MyDocument < OSX::NSDocument
	ib_outlets :window
	ib_outlets :transcriptView, :importXMLButton, :copyClipsButton, :transcriptSpinner1, :clipsTableView, :colorPopupMenu, :subclipNameField, :thumbnailImageView, :reconnectMediaButton, :goodCheckbox1, :priorityPopupMenu, :descriptionComboBox, :sceneComboBox, :shottakeComboBox, :lognoteComboBox, :keywordsTokenField, :locationComboBox, :peopleTokenField, :eventComboBox, :brollKeywordsComboBox, :commentbComboBox, :transcriptSearchField, :transcriptSpinner2, :transcriptPreviewButton, :transcriptTableView, :transcriptJoinButton
	ib_outlets :storyView, :searchPopupMenu, :selectsSearchField, :selectsSpinner, :selectsPreviewButton, :selectsTableView, :goodCheckbox2, :addButton, :storyBox, :showThumbnailsCheckbox, :showColorsCheckbox, :storySpinner, :storyPreviewButton, :storyTableView, :removeButton, :storyJoinButton, :addVOButton, :deselectAllButton, :exportSpinner, :exportButton
	ib_outlets :progressWindow, :progressView, :progressIndicator, :progressMessage
	ib_outlets :clipCopyDialog, :projectsPopup
	ib_outlets :exportDialog, :frameRatePopup, :sendToPopup
	ib_outlets :regoView, :deactivateAreaView, :webView, :webSpinner, :problemMessage, :closeRegoButton

	def showTranscriptView(sender)
		#puts "showTranscriptView"
		@project_array[0] = TOOLBAR_FIRST_IDENTIFIER
		populate_transcript_view()
		if @subclips_array.empty?
			@importXMLButton.setEnabled(true)
		else
			@importXMLButton.setEnabled(false)
		end
		unless @current_view==@transcriptView
			@transcriptView.setFrame(@storyView.frame)
			@toolbar.selectedItemIdentifier = @toolbar.items[0].itemIdentifier
			@clips_table_view_array, @transcript_table_view_array = transcript_view_data_sources()
			if @is_Lion # FIXME
				@documentWindow.contentView.replaceSubview_with(@current_view, @transcriptView)
			else
				@documentWindow.contentView.setWantsLayer(true)
				@documentWindow.contentView.displayIfNeeded
				transition = OSX::CATransition.animation
				transition.setDelegate(self)
				transition.setType(KCATransitionPush)
				transition.setSubtype(KCATransitionFromLeft)
				# transition.setDuration(0.5)
				ani = OSX::NSDictionary.dictionaryWithObject_forKey(transition, "subviews")
				@documentWindow.contentView.setAnimations(ani)
				@documentWindow.contentView.animator.replaceSubview_with(@current_view, @transcriptView)
			end
			@current_view = @transcriptView
		end
	end
	ib_action :showTranscriptView

	def showStoryView(sender)
		#puts "showStoryView"
		@project_array[0] = TOOLBAR_SECOND_IDENTIFIER
		@selects_table_view_array, @story_table_view_array, @clips_table_view_array = story_view_data_sources()
		#puts "@selects_table_view_array.length = #{@selects_table_view_array.length}"
		@selectsTableView.reloadData
		reorderStoryList() # calculate Story duration
		showThumbnails(self) # also does @storyTableView.reloadData
		unless @current_view==@storyView
			@storyView.setFrame(@transcriptView.frame)
			@toolbar.selectedItemIdentifier = @toolbar.items[1].itemIdentifier
			if @is_Lion # FIXME
				@documentWindow.contentView.replaceSubview_with(@current_view, @storyView)
			else
				@documentWindow.contentView.setWantsLayer(true)
				@documentWindow.contentView.displayIfNeeded
				transition = OSX::CATransition.animation
				transition.setDelegate(self)
				transition.setType(KCATransitionPush)
				transition.setSubtype(KCATransitionFromRight)
				# transition.setDuration(0.5)
				ani = OSX::NSDictionary.dictionaryWithObject_forKey(transition, "subviews")
				@documentWindow.contentView.setAnimations(ani)
				@documentWindow.contentView.animator.replaceSubview_with(@current_view, @storyView)
			end
			@current_view = @storyView
		end
		save_all_fields()
	end
	ib_action :showStoryView

	def animationDidStop_finished(theAnimation, bool)
		if bool==true
			@documentWindow.contentView.setWantsLayer(false)
		end
	end

	def initialize
		#puts "initialize"
		@xmpcommand = OSX::NSBundle.mainBundle.resourcePath.fileSystemRepresentation + "/third-party/xmpcommand"
		@black_color, @light_gray_color, @white_color = OSX::NSColor.blackColor, OSX::NSColor.lightGrayColor, OSX::NSColor.whiteColor
		blue = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(220.0/255.0, 240.0/255.0, 255.0/255.0, 1.0)
		green = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(230.0/255.0, 255.0/255.0, 230.0/255.0, 1.0)
		lavender = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(230.0/255.0, 230.0/255.0, 255.0/255.0, 1.0)
		lemon = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(255.0/255.0, 255.0/255.0, 230.0/255.0, 1.0)
		lime = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(240.0/255.0, 255.0/255.0, 220.0/255.0, 1.0)
		orange = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(255.0/255.0, 240.0/255.0, 220.0/255.0, 1.0)
		pink = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(255.0/255.0, 230.0/255.0, 255.0/255.0, 1.0)
		purple = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(240.0/255.0, 220.0/255.0, 255.0/255.0, 1.0)
		salmon = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(255.0/255.0, 230.0/255.0, 230.0/255.0, 1.0)
		sea = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(230.0/255.0, 255.0/255.0, 255.0/255.0, 1.0)
		sunset = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(255.0/255.0, 220.0/255.0, 240.0/255.0, 1.0)
		teal = OSX::NSColor.colorWithCalibratedRed_green_blue_alpha(220.0/255.0, 255.0/255.0, 240.0/255.0, 1.0)
		white = OSX::NSColor.whiteColor
		@bg_colors = [white, nil, blue, green, lavender, lemon, lime, orange, pink, purple, salmon, sea, sunset, teal]
		resources = OSX::NSBundle.mainBundle.resourcePath.fileSystemRepresentation
		@default_image = OSX::NSImage.alloc.initByReferencingFile(resources + "/NSUser.tiff")
		@offline_image = OSX::NSImage.alloc.initByReferencingFile(resources + "/Offline.tiff")
		@voiceover_image = OSX::NSImage.alloc.initByReferencingFile(resources + "/VoiceOver.tiff")
		@project_array = [ TOOLBAR_FIRST_IDENTIFIER, "", "", [], 0, 0] 
		@clips_table_view_array, @transcript_table_view_array, @selects_table_view_array, @story_table_view_array = [], [], [], []
		@maximum_handle_duration = 0.5 * 1000 # 0.5 secs is 500 ms handle - could set this from a preferences gadget
		@players = Array.new
		values = OSX::NSUserDefaultsController.sharedUserDefaultsController.values
		if values.valueForKey('Autosave').nil? # set default Autosave
			defaults = OSX::NSUserDefaults.standardUserDefaults
			defaults.setObject_forKey(1800, 'Autosave')
			defaults.synchronize # saves the plist
		end
		check_for_expired_beta()
		if @this_rego[0]["active"]
			OSX::NSDocumentController.sharedDocumentController.setAutosavingDelay( values.valueForKey('Autosave') ) # time interval in seconds for periodic autosaving
		else
			OSX::NSDocumentController.sharedDocumentController.setAutosavingDelay(0)
		end
		#puts "Autosave every #{values.valueForKey('Autosave')} seconds"
		#puts "Is this Lion?"
		nsAppKitVersionNumber10_7 = 1138
		nsAppKitVersionNumber10_8 = 1187
		# if OSX::NSAppKitVersionNumber.floor < nsAppKitVersionNumber10_7 || OSX::NSAppKitVersionNumber.floor > nsAppKitVersionNumber10_8
			# @is_Lion = false # Snow Leopard or Mavericks
		# else
			@is_Lion = true
		# end
		#puts "is_Lion #{@is_Lion}"
	end

	def windowNibName
		"MyDocument"
	end

	def windowControllerDidLoadNib(aController)
		#puts "windowControllerDidLoadNib"
		super_windowControllerDidLoadNib(aController)
		# Add any code here that need to be executed once the
		# windowController has loaded the document's window.
		check_for_expired_beta()
		@documentWindow = aController.window
		############# NEW CODE
		# contentView = @documentWindow.contentView
		#     contentView.setWantsLayer(true)
		#     transition = OSX::CATransition.animation
		#     transition.setType(KCATransitionPush)
		#     transition.setSubtype(KCATransitionFromRight)
		# ani = OSX::NSDictionary.dictionaryWithObject_forKey(transition, "subviews")
		# contentView.animations = ani
		#############
		@documentWindow.contentView.addSubview(@transcriptView)
		setupToolbar()
		@toolbar.selectedItemIdentifier = @toolbar.items[0].itemIdentifier
		@tokenFieldCell = OSX::NSTokenFieldCell.new
		@textFieldCell = @selectsTableView.tableColumns[1].dataCell
		@imageCell = OSX::NSImageCell.new
		@selectsTableView.tableColumns[0].setDataCell(@tokenFieldCell)
		@selectsTableView.registerForDraggedTypes([StoryPasteboard])
		@storyTableView.registerForDraggedTypes([SelectsPasteboard, StoryPasteboard])
		@webView.setDrawsBackground(false)
		@current_view = @transcriptView
		if @this_rego[0]["active"]
			if @project_array[0]==TOOLBAR_FIRST_IDENTIFIER
				showTranscriptView(self)
			elsif @project_array[0]==TOOLBAR_SECOND_IDENTIFIER
				showStoryView(self)
				@documentWindow.contentView.setWantsLayer(false)
			end
		elsif @this_rego[0]["mac_problem"]==true
			check_for_rego(@this_rego)
		else
			# @deactivateMenuItem.setTitle("Activate...")
			displayDemoSheet()
		end
		pathurl = OSX::NSBundle.mainBundle.resourcePath.fileSystemRepresentation + "/slugs/720x480.mov"
		qt = QuickTime::Movie.empty # rmov
		getThumbnails unless @thumbnailed==true
		qt.dispose
	end

	### REGO METHODS ###
	
	def displayDemoSheet
		alert = OSX::NSAlert.alloc.init
		alert.setMessageText("#{SKU} has not been activated for this computer.")
		alert.setAlertStyle(OSX::NSInformationalAlertStyle)
		alert.setInformativeText("Would you like to activate #{SKU}, or use it in demo mode (which is unable to save or export)?")
		alert.addButtonWithTitle("Activate")
		alert.addButtonWithTitle("Buy")
		alert.addButtonWithTitle("Demo")
		alert.beginSheetModalForWindow_modalDelegate_didEndSelector_contextInfo(@mainWindow, self, "didEndDemoSheet:returnCode:contextInfo:", nil)
	end

	def didEndDemoSheet_returnCode_contextInfo(sheet, buttonClicked, info)
		sheet.window.orderOut(nil)
		if buttonClicked==OSX::NSAlertSecondButtonReturn
			system('open', 'http://www.AssistedEditing.com/prEdit/buy')
		end
		if buttonClicked==OSX::NSAlertThirdButtonReturn
			showTranscriptView(self)
		else
			showActivationView(nil, WEB_APP + "license/" + SKU.gsub(" ", "%20") + "?sn=" + @this_rego[0]["number"].to_s + "&mac=" + MAC)
		end
	end

	def showDeactivation
		# if @deactivateMenuItem.title=="Activate..."
		# 	showActivationView(newFrame, WEB_APP + "license/" + SKU.gsub(" ", "%20") + "?sn=" + @this_rego[0]["number"].to_s + "&mac=" + MAC) # NEW
		# else
		if @this_rego[0]["active"]
			showActivationView(nil, WEB_APP + "destroy/" + @this_rego[0]["number"].to_s)
		end
		@closeRegoButton.setTitle("Cancel")
	end
	ib_action :showDeactivation # prEdit > Activate or Deactivate
	
	def showActivationView(frame, url)
		#puts "showActivationView #{url}", @current_view.inspect
		@documentWindow.contentView.replaceSubview_with(@current_view, @regoView)
		@documentWindow.setTitle("Activation")
		@current_view = @regoView
		# notification
		center = OSX::NSNotificationCenter.defaultCenter
		center.addObserver_selector_name_object(self, :webProgress, nil, @webView)
		@webView.mainFrame.loadRequest( OSX::NSURLRequest.requestWithURL( OSX::NSURL.URLWithString(url) ) )
	end
	
	def webProgress(notification)
		case notification.name
		when "WebProgressStartedNotification"
			@webSpinner.startAnimation(self)
		when "WebProgressFinishedNotification"
			@webSpinner.stopAnimation(self)
			if @webView.mainFrame.dataSource.nil? # in case page doesn't load...
				@webView.mainFrame.stopLoading
				@problemMessage.setHidden(false)
				@webView.setHidden(true)
			else
				get_web_app_response(@webView.mainFrame.dataSource.response.URL)
			end
		end
	end 
	
	def closeActivationView
		if @closeRegoButton.title=="Continue" || @closeRegoButton.title=="Cancel"
			@documentWindow.setTitle("Untitled")
			showTranscriptView(self)
		else
			OSX::NSApplication.sharedApplication.terminate(self)
		end
	end
	ib_action :closeActivationView # on click of Registration window's Continue button
	
	### end rego methods ###

	def showHelp(sender)
		#puts "showHelp"
		openPDF("prEdit - Help.pdf")
	end
	ib_action :showHelp # Help > prEdit Help
	
	def openPDF(filename)
		path_to_file = __FILE__.sub("MyDocument.rb",filename)
		system('open', path_to_file)
	end
	
	def visitWebsite(sender)
		system('open', 'http://www.assistedediting.com/')
	end
	ib_action :visitWebsite # Help > Visit www.assistedediting.com
	
	def sendFeedback(sender)
		OSX::JRFeedbackController.showFeedback
	end
	ib_action :sendFeedback # prEdit > Provide Feedback

	def sendSupportEmail(subject)
		system('open', "mailto:techsupport@intelligentassistance.com?subject=prEdit #{subject}&body=Attached is the XML file(s) that Transcriptize was unable to process.")
	end

	### COMMON METHODS ###
	
	def writeToURL_ofType_error(absoluteURL, typeName, error)
		@transcriptSpinner1.startAnimation(self)
		@exportSpinner.startAnimation(self)
		begin
			result = save_the_project(absoluteURL.path)
		rescue
			result = false
		end
		@transcriptSpinner1.stopAnimation(self)
		@exportSpinner.stopAnimation(self)
		return result
	end
	
	def readFromURL_ofType_error(absoluteURL, typeName, error)
		begin
			#puts "it's Marshal, not YAML"
			file = File.open(absoluteURL.path, 'r')
			@project_array = Marshal.load file.read
			file.close
		rescue
			#puts "it's YAML, not Marshal"
			@project_array = YAML.load_file(absoluteURL.path)
		end
		# convert string keys to symbols for speed
		#puts "start conversion of string keys to symbols..."
		array_with_string_keys = @project_array[3]
		array_with_symbol_keys = Array.new
		array_with_string_keys.each do |hash|
			hash_with_symbol_keys = Hash.new
			hash.each_pair do |key, value|
				if key==:thumbnail
					hash_with_symbol_keys[key] = nil
				else
					hash_with_symbol_keys[key.to_sym] = value
				end
			end
			array_with_symbol_keys << hash_with_symbol_keys
		end
		# convert xml string to array of clips
		if @project_array[2].class==String
			convert_xml_to_array_of_clips()
		end
		# repair broken :timestamps_array and build :searchable_transcript
		array_with_symbol_keys.each_with_index do |subclip, i|
			unless subclip[:timestamps_array][0].class==Array
				first_timestamp = [ subclip[:timestamps_array][0], subclip[:timestamps_array][1] ]
				subclip[:timestamps_array][0] = first_timestamp
				subclip[:timestamps_array].delete_at(1)
				if subclip[:in_point]==0 && i!=0 && subclip[:clip_id]==array_with_symbol_keys[i-1][:clip_id]
					in_point = first_word_with_handle(subclip[:timestamps_array][0][0], array_with_symbol_keys[i-1][:timestamps_array][-1][-1]) # change the next row's in_point
					#puts "...changing clip #{i+1} :in_point from 0 to #{in_point}"
					subclip[:in_point] = in_point
				end
			end
			subclip[:searchable_transcript] = subclip[:Transcript].gsub("_"," ").downcase
		end # end repair
		@project_array[3] = array_with_symbol_keys
		@subclips_array = @project_array[3]
		# enable_lognotes_and_comments(true, [0])
		# @thumbnailImageView.setNeedsDisplay(true)
		#puts "...done"
		return true
	end
	
	def getThumbnails
		unless @subclips_array.nil? || @subclips_array.empty?
			@transcriptSpinner1.startAnimation(self)
			@exportSpinner.startAnimation(self)
			begin
				@thumbs = Thread.new do # Thread is better than NSThread because it doesn't block the GUI
					#puts "Making thumbnails..."
					@subclips_array.each do |subclip|
						thumbnail_path = THUMBNAILS_FOLDER + subclip[:clip_name] + "/" + subclip[:in_point].to_i.to_s + ".png"
						if subclip[:thumbnail].nil? # only make new thumbnails (allows us to quickly go back through the getThumbnails method)
							if File.file?(subclip[:pathurl])==false # offline file
								subclip[:thumbnail] = @offline_image
							elsif subclip[:clip_name]=="Voice Overs" || subclip[:width].nil? # audio-only files
								subclip[:thumbnail] = @voiceover_image
							elsif File.file?(thumbnail_path)==false # no thumbnail saved yet
								array_of_subclips = @subclips_array.find_all { |s| s[:clip_id] == subclip[:clip_id] } # clip_id uniquely identifies the media clip
								create_mov_thumbnails(array_of_subclips, 0.2) # sleep time
							else # load the image thumbnail
								subclip[:thumbnail] = OSX::NSImage.alloc.initByReferencingFile(thumbnail_path) # can be nil if this fails
							end
						end
					end # end @subclips_array.each
					@thumbnailed = true
					#puts "...done making thumbnails!"
				end # end Thread.new
			rescue
				@thumbnailed = true # false
				#puts "...can't finish making thumbnails!"
			end
			@transcriptSpinner1.stopAnimation(self)
			@exportSpinner.stopAnimation(self)
			@thumbnailImageView.setNeedsDisplay(true)
			@reconnectMediaButton.setEnabled(true) unless @subclips_array.empty?
			@storyTableView.setNeedsDisplay(true) if @showThumbnailsCheckbox.state==1 # redisplays @storyTableView without reloadData
		end
	end

	# splits both :Transcript and :timestamps_array, and reorders :subclip_id, in table_view_array and @subclips_array
	def splitRow(params)
		#puts "splitRow"
		tag, edited_row, string = params["tag"].to_i, params["row"].to_i, params["string"].to_s
		case tag
		when 2
			@tableView = @transcriptTableView
			@table_view_array = @transcript_table_view_array
		when 4
			@tableView = @storyTableView
			@table_view_array = @story_table_view_array
		end
		split_row(edited_row, string)
		@tableView.reloadData # redisplays the new row heights
		# add to undoManager
		undoManager.registerUndoWithTarget_selector_object(self, 'joinRows', {"tag" => @tableView.tag, "rows" => [edited_row, edited_row+1] } )
		undoManager.setActionName("Subclip Edit")
	end

	def joinRows(params)
		#puts "joinRows"
		tag, rows = params["tag"].to_i, params["rows"].to_a
		case tag
		when 2
			@tableView = @transcriptTableView
			@table_view_array = @transcript_table_view_array
		when 4
			@tableView = @storyTableView
			@table_view_array = @story_table_view_array
		end
		first_row = rows[0].to_i
		undo_array = join_rows(rows)
		@tableView.reloadData
		@tableView.deselectAll(self)
		reorderStoryList()
		# add to undoManager
		undoManager.registerUndoWithTarget_selector_object(self, 'splitRow', {"tag" => @tableView.tag, "row" => first_row, "string" => undo_array.join("\n") } )
		undoManager.setActionName("Subclip Edit")
	end

	def showCopyClipsDialog(sender)
		@openProjects = OSX::NSDocumentController.sharedDocumentController.documents
		@projectsPopup.removeAllItems
		@projectsPopup.setAutoenablesItems(false)
		@openProjects.each do |p|
			@projectsPopup.addItemWithTitle(p.displayName)
			if p == OSX::NSDocumentController.sharedDocumentController.currentDocument
				@projectsPopup.itemArray[-1].setEnabled(false) # gray out this project
			end
		end
		OSX::NSApp.beginSheet_modalForWindow_modalDelegate_didEndSelector_contextInfo( @clipCopyDialog, @documentWindow, self, 'didEndClipCopy:returnCode:contextInfo:', nil )
	end
	ib_action :showCopyClipsDialog
	
	def closeCopyClipsDialog(sender)
		#puts "closeCopyClipsDialog"
		@clipCopyDialog.orderOut(nil)
		case sender.tag
		when 0 # Cancel
			OSX::NSApp.endSheet_returnCode(@clipCopyDialog, 0)
		when 1 # OK
			OSX::NSApp.endSheet_returnCode(@clipCopyDialog, 1)
		end
	end
	ib_action :closeCopyClipsDialog
	
	def didEndClipCopy_returnCode_contextInfo(sheet, code, info)
		#puts "didEndClipCopy_returnCode_contextInfo"
		if code==1
			#puts "OK"
			to_project = @openProjects[@projectsPopup.indexOfSelectedItem]
			to_project.add_clips_to_self(@subclips_array.dup, @project_array[2].dup, @xml_name)
		elsif code==0
			#puts "Cancel"
		end
	end
	
	# called by didEndClipCopy_returnCode_contextInfo in another MyDocument
	def add_clips_to_self(other_subclips_array, other_array_of_clips, other_xml_name)
		@transcriptSpinner1.startAnimation(self)
		if @subclips_array.empty?
			other_subclips_array.each do |hash|
				hash[:added_to_story] = false
				hash[:story_order] = -1
			end
			@project_array[3] = @subclips_array = other_subclips_array # data source of all subclips
			@project_array[2] = other_array_of_clips # array of clip nodes
			@project_array[1] = @xml_name = other_xml_name
		else
			last_clip_id = @subclips_array[-1][:clip_id]
			#puts "last_clip_id = #{last_clip_id}"
			other_subclips_array.each do |hash|
				hash[:clip_id] = hash[:clip_id] + 1 + last_clip_id
				#puts "hash[:clip_id] = #{hash[:clip_id]}"
				hash[:added_to_story] = false
				hash[:story_order] = -1
				@project_array[3] << hash
			end
			@project_array[2] = @project_array[2] + other_array_of_clips
			#puts @project_array[2].length
		end
		@transcriptSpinner1.stopAnimation(self)
		showTranscriptView(self)
		self.updateChangeCount(OSX::NSChangeDone)
	end
	
	def displayAlert(title, body, style=0)
		# 0 = NSInformationalAlertStyle, 1 = NSWarningAlertStyle, 2 = NSCriticalAlertStyle (has caution icon)
		alert = OSX::NSAlert.alloc.init
		alert.setMessageText(title)
		alert.setAlertStyle(style) 
		alert.setInformativeText(body)
		alert.runModal
	end
	
	### END COMMON METHODS ###
	
	### TRANSCRIPT VIEW METHODS ###

	def chooseXmlFile(sender)
		filetypes = ["xml"]
		oPanel = OSX::NSOpenPanel.openPanel
		oPanel.setAllowsMultipleSelection(false)
		oPanel.beginSheetForDirectory_file_types_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, filetypes, @documentWindow, self, 'didEndXmlOpenSheet:returnCode:contextInfo:', nil)
	end
	ib_action :chooseXmlFile

	def didEndXmlOpenSheet_returnCode_contextInfo(oPanel, buttonClicked, info)
		if buttonClicked == OSX::NSOKButton
			oPanel.orderOut(self)
			@transcriptSpinner1.startAnimation(self)
			path = oPanel.URL.path.to_s
			begin
				get_info_from_clips(path.to_s)
				populate_transcript_view()
				getThumbnails
				self.updateChangeCount(OSX::NSChangeDone)
			rescue
				@transcriptSpinner1.stopAnimation(self)
				displayAlert("There was a problem getting information about these clips.", "Please choose another XML file exported from\nFinal Cut Pro or Premiere Pro.")
				#puts "PROBLEM with get_info_from_clips()"
			end
		end
	end

	def setPaperColor(sender)
		#puts "setPaperColor"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		selected_rows.each do |row|
			transcriptCell = @transcriptTableView.tableColumns[0].dataCellForRow(row)
			transcriptCell.setBackgroundColor( @bg_colors[@colorPopupMenu.indexOfSelectedItem] )
			@transcript_table_view_array[row][:bg_color_index] = @colorPopupMenu.indexOfSelectedItem
		end
		@transcriptTableView.deselectAll(self)
		self.updateChangeCount(OSX::NSChangeDone)
	end
	ib_action :setPaperColor

	def reconnectMedia(sender)
		offline_clip = @subclips_array.find { |subclip| subclip[:thumbnail] == @offline_image }
		#puts "is offline_clip.nil? #{offline_clip.nil?}"
		if offline_clip.nil?
			@reconnect_pathurl = @subclips_array[0][:pathurl]
		else
			@reconnect_pathurl = offline_clip[:pathurl]
		end
		oPanel = OSX::NSOpenPanel.openPanel
		oPanel.setDelegate(self) # uses delegate method panel_shouldShowFilename
		oPanel.setAllowsMultipleSelection(false)
		oPanel.setMessage("Choose the new location of #{File.basename(@reconnect_pathurl)}")
		oPanel.setPrompt("Choose")
		oPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, nil, @documentWindow, self, 'didEndReconnectSheet:returnCode:contextInfo:', nil)
	end
	
	def didEndReconnectSheet_returnCode_contextInfo(oPanel, buttonClicked, info)
		if buttonClicked == OSX::NSOKButton
			oPanel.orderOut(self)
			@transcriptSpinner1.startAnimation(self)
			path = oPanel.URL.path.to_s
			reconnect_pathurl_array = @reconnect_pathurl.split("/")
			reconnect_pathurl_rev_array = @reconnect_pathurl.split("/").reverse!
			new_pathurl_array = path.to_s.split("/")
			new_pathurl_rev_array = path.split("/").reverse!
			stop_loop = false
			new_pathurl_rev_array.each_index do |i|
				if new_pathurl_rev_array[i]==reconnect_pathurl_rev_array[i] && stop_loop==false
					reconnect_pathurl_array.delete_at(-1)
					new_pathurl_array.delete_at(-1)
				else
					stop_loop = true
				end
			end
			reconnected_clips = 0
			replace_filepath = reconnect_pathurl_array.join("/")
			with_filepath = new_pathurl_array.join("/")
			@subclips_array.each do |subclip|
				new_pathurl = subclip[:pathurl].sub(replace_filepath, with_filepath)
				if subclip[:pathurl]==new_pathurl
					#puts "pathurl hasn't changed"
				elsif File.exist?(new_pathurl)
					oldNSURL = OSX::NSURL.fileURLWithPath(subclip[:pathurl])
					newNSURL = OSX::NSURL.fileURLWithPath(new_pathurl)
					subclip[:pathurl] = new_pathurl
					#puts "changed pathurl to #{new_pathurl}"
					@project_array[2].each do |clip_string| # array of clips XML
						# find oldNSURL string and replace with newNSURL string
						clip_string.sub!(oldNSURL.absoluteString, newNSURL.absoluteString)
					end
					create_mov_thumbnails([subclip], 0)
					reconnected_clips += 1
				else
					#puts "missing file - leave pathurl unchanged"
				end
			end
			#puts "reconnected #{reconnected_clips} subclips"
			unless reconnected_clips==0
				self.updateChangeCount(OSX::NSChangeDone)
				selected_rows = @transcriptTableView.selectedRowIndexes.to_a
				if selected_rows.length==1
					enable_lognotes_and_comments(true, selected_rows) # reloads the @thumbnailImageView if it's changed
				end
			end
		end
		@transcriptSpinner1.stopAnimation(self)
	end
	ib_action :reconnectMedia
	
	# editField is sent params for the field, selected row, new value and Action Name; it also registers the old value with undoManager
	def editField(params)
		#puts "editField"
		selected_row, new_value, action_name = params["selected_row"].to_i, params["value"].to_s, params["action_name"].to_s
		field = params["field"].to_sym
		current_value = @transcript_table_view_array[selected_row][field]
		if new_value!=current_value
			# add to undoManager
			undoManager.registerUndoWithTarget_selector_object(self, 'editField', {"selected_row" => selected_row, "field" => field, "value" => current_value, "action_name" => action_name } )
			undoManager.setActionName(action_name)
			@transcript_table_view_array[selected_row][field] = new_value
			enable_lognotes_and_comments(true, [selected_row])
			#puts @transcript_table_view_array[selected_row].inspect
		end
	end
	
	def setSubclipName(sender)
		#puts "setSubclipName"
		selected_row = @transcriptTableView.selectedRowIndexes.to_a[0]
		selected_row = 0 if selected_row.nil?
		new_value = @subclipNameField.stringValue.to_s
		editField( {"selected_row" => selected_row, "field" => :subclip_name, "value" => new_value, "action_name" => "Subclip Name Change" } )
	end
	ib_action :setSubclipName

	def markGood(sender)
		#puts "markGood"
		if sender.tag==1 # Transcript view checkbox (allows mixed state)
			selectedRowIndexes = @transcriptTableView.selectedRowIndexes
			selectedRowIndexes.to_a.each do |row|
				if @goodCheckbox1.state==0
					# @transcript_table_view_array[row][:good] = "FALSE"
					editField( {"selected_row" => row, "field" => :good, "value" => "FALSE", "action_name" => "Good Change" } )
				elsif @goodCheckbox1.state==1
					# @transcript_table_view_array[row][:good] = "TRUE"
					editField( {"selected_row" => row, "field" => :good, "value" => "TRUE", "action_name" => "Good Change" } )
				else # mixed state
					@goodCheckbox1.setState(1)
					editField( {"selected_row" => row, "field" => :good, "value" => "TRUE", "action_name" => "Good Change" } )
				end
			end
			@transcriptTableView.reloadData
			@transcriptTableView.selectRowIndexes_byExtendingSelection( selectedRowIndexes, false) # If the extend flag is NO the selected rows are specified by indexes. If extend is YES, the rows indicated by indexes are added to the collection of already selected rows, providing multiple selection.
		else # Selects list checkbox (on by default, no mixed state)
			altered_rows = Array.new
			@selectsTableView.selectedRowIndexes.to_a.reverse_each do |r|
				altered_rows << {"clip_id" => @selects_table_view_array[r][:clip_id], "subclip_id" => @selects_table_view_array[r][:subclip_id], "good" => @selects_table_view_array[r][:good].to_s, "added_to_story" => @selects_table_view_array[r][:added_to_story].to_s, "story_order" => @selects_table_view_array[r][:story_order].to_s} 
				@selects_table_view_array[r][:good] = "FALSE"
				@selects_table_view_array.delete_at(r) # remove from selects list
			end
			@selectsTableView.reloadData
			@selectsTableView.deselectAll(self)
			# add to undoManager
			undoManager.registerUndoWithTarget_selector_object(self, 'changeSubclipsArray', {"undo_array" => altered_rows, "action_name" => "Good Change"} )
			undoManager.setActionName("Good Change")
		end
	end
	ib_action :markGood

	def setPriority(sender)
		#puts "setPriority"
		selected_row = @transcriptTableView.selectedRowIndexes.to_a[0]
		selected_row = 0 if selected_row.nil?
		new_value = @priorityPopupMenu.indexOfSelectedItem
		editField( {"selected_row" => selected_row, "field" => :priority, "value" => new_value, "action_name" => "Priority Change" } )
	end
	ib_action :setPriority

	def addDescription(sender)
		#puts "addDescription"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @descriptionComboBox.stringValue.to_s
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:description]
				editField( {"selected_row" => row, "field" => :description, "value" => new_value, "action_name" => "Description Change" } )
				@descriptions << new_value
			end
		end
		@descriptions = @descriptions.uniq.sort
		@descriptionComboBox.removeAllItems
		@descriptionComboBox.addItemsWithObjectValues(@descriptions)
	end
	ib_action :addDescription

	def addScene(sender)
		#puts "addScene"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @sceneComboBox.stringValue.to_s
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:scene]
				editField( {"selected_row" => row, "field" => :scene, "value" => new_value, "action_name" => "Scene Change" } )
				@scenes << new_value
			end
		end
		@scenes = @scenes.uniq.sort
		@sceneComboBox.removeAllItems
		@sceneComboBox.addItemsWithObjectValues(@scenes)
	end
	ib_action :addScene

	def incrementScene(sender)
		#puts "incrementScene"
		if @sceneComboBox.isEnabled
			current_value = @sceneComboBox.stringValue.to_s
			@sceneComboBox.setStringValue(current_value.succ)
			addScene(self) 
		end
	end
	ib_action :incrementScene

	def addShotTake(sender)
		#puts "addShotTake"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @shottakeComboBox.stringValue.to_s
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:shottake]
				editField( {"selected_row" => row, "field" => :shottake, "value" => new_value, "action_name" => "Shot/Take Change" } )
				@shottakes << new_value
			end
		end
		@shottakes = @shottakes.uniq.sort
		@shottakeComboBox.removeAllItems
		@shottakeComboBox.addItemsWithObjectValues(@shottakes)
	end
	ib_action :addShotTake

	def incrementShotTake(sender)
		#puts "incrementShotTake"
		if @shottakeComboBox.isEnabled
			current_value = @shottakeComboBox.stringValue.to_s
			@shottakeComboBox.setStringValue(current_value.succ)
			addShotTake(self) 
		end
	end
	ib_action :incrementShotTake

	def addLogNote(sender)
		#puts "addLogNote"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @lognoteComboBox.stringValue.to_s
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:lognote]
				editField( {"selected_row" => row, "field" => :lognote, "value" => new_value, "action_name" => "Log Note Change" } )
				@lognotes << new_value
			end
		end
		@lognotes = @lognotes.uniq.sort
		@lognoteComboBox.removeAllItems
		@lognoteComboBox.addItemsWithObjectValues(@lognotes)
	end
	ib_action :addLogNote

	def addKeywords(sender)
		#puts "addKeywords"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @keywordsTokenField.stringValue.to_s.sub(", ?","").sub(",?","") # clean off stray questionmarks
		@keywordsTokenField.setStringValue(new_value)
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:mastercomment1]
				editField( {"selected_row" => row, "field" => :mastercomment1, "value" => new_value, "action_name" => "Story Keywords Change" } )
				new_values = new_value.split(",").each { |n| n.strip! }
				@keywords += new_values
			end
		end
		@keywords = @keywords.uniq.sort
		#puts "@keywords #{@keywords.inspect}"
	end
	ib_action :addKeywords

	def addLocation(sender)
		#puts "addLocation"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @locationComboBox.stringValue.to_s
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:mastercomment2]
				editField( {"selected_row" => row, "field" => :mastercomment2, "value" => new_value, "action_name" => "Location Change" } )
				@locations << new_value
			end
		end
		@locations = @locations.uniq.sort
		@locationComboBox.removeAllItems
		@locationComboBox.addItemsWithObjectValues(@locations)
	end
	ib_action :addLocation

	def addPeople(sender)
		#puts "addPeople", @peopleTokenField.stringValue.inspect
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @peopleTokenField.stringValue.to_s.sub(", ?","").sub(",?","") # clean off stray questionmarks
		@peopleTokenField.setStringValue(new_value)
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:Person]
				editField( {"selected_row" => row, "field" => :Person, "value" => new_value, "action_name" => "Person Change" } )
				new_values = new_value.split(",").each { |n| n.strip! }
				@people += new_values
			end
		end
		@people = @people.uniq.sort
		#puts "@people #{@people.inspect}"
	end
	ib_action :addPeople

	def addEvent(sender)
		#puts "addEvent"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @eventComboBox.stringValue.to_s
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:mastercomment4]
				editField( {"selected_row" => row, "field" => :mastercomment4, "value" => new_value, "action_name" => "Event or Theme Change" } )
				@events << new_value
			end
		end
		@events = @events.uniq.sort
		@eventComboBox.removeAllItems
		@eventComboBox.addItemsWithObjectValues(@events)
	end
	ib_action :addEvent

	def addBrollKeyword(sender)
		#puts "addBrollKeyword"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @brollKeywordsComboBox.stringValue.to_s
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:clipcommenta]
				editField( {"selected_row" => row, "field" => :clipcommenta, "value" => new_value, "action_name" => "Comment A Change" } )
				@broll_keywords << new_value
			end
		end
		@broll_keywords = @broll_keywords.uniq.sort
		@brollKeywordsComboBox.removeAllItems
		@brollKeywordsComboBox.addItemsWithObjectValues(@broll_keywords)
	end
	ib_action :addBrollKeyword

	def addCommentB(sender)
		#puts "addCommentB"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		new_value = @commentbComboBox.stringValue.to_s
		selected_rows.each do |row|
			if new_value!=@transcript_table_view_array[row][:clipcommentb]
				editField( {"selected_row" => row, "field" => :clipcommentb, "value" => new_value, "action_name" => "Comment B Change" } )
				@comment_bs << new_value
			end
		end
		@comment_bs = @comment_bs.uniq.sort
		@commentbComboBox.removeAllItems
		@commentbComboBox.addItemsWithObjectValues(@comment_bs)
	end
	ib_action :addCommentB

	def save_all_fields
		if @transcriptTableView.selectedRowIndexes.to_a.length > 0
			setSubclipName(self) if @subclipNameField.isEnabled
			addDescription(self) if @descriptionComboBox.isEnabled
			addScene(self) if @sceneComboBox.isEnabled
			addShotTake(self) if @shottakeComboBox.isEnabled
			addLogNote(self) if @lognoteComboBox.isEnabled
			addKeywords(self) if @keywordsTokenField.isEnabled
			addLocation(self) if @locationComboBox.isEnabled
			addPeople(self) if @peopleTokenField.isEnabled
			addEvent(self) if @eventComboBox.isEnabled
			addBrollKeyword(self) if @brollKeywordsComboBox.isEnabled
			addCommentB(self) if @commentbComboBox.isEnabled
		end
	end
	
	def filterTranscripts(sender)
		filter_string = @transcriptSearchField.stringValue.to_s.downcase
		#puts "filterTranscripts: search for #{filter_string.inspect}"
		if filter_string=="" # cancel
			@clips_table_view_array, @transcript_table_view_array = transcript_view_data_sources()
		else
			transcript_table_view_array_copy = @subclips_array.find_all { |s| s[:clip_id]==@selected_clip_index }
			@transcript_table_view_array = []
			case @searchPopupMenu.titleOfSelectedItem
			when "Everywhere"
				#puts "Everywhere"
				transcript_table_view_array_copy.each do |params|
					content_string = params[:Person].downcase + " " + params[:searchable_transcript] + " " + " " + params[:mastercomment1].downcase + params[:mastercomment2].downcase + " " + params[:mastercomment4].downcase
					@transcript_table_view_array << params unless content_string.index(filter_string).nil? # does contain filter_string?
				end
			when "Transcript"
				#puts "Transcript"
				transcript_table_view_array_copy.each do |params|
					@transcript_table_view_array << params unless params[:searchable_transcript].index(filter_string).nil? # does contain filter_string?
				end
			when "Story Keywords"
				#puts "Story Keywords"
				transcript_table_view_array_copy.each do |params|
					@transcript_table_view_array << params unless params[:mastercomment1].downcase.index(filter_string).nil? # does contain filter_string?
				end
			when "Location"
				#puts "Location"
				transcript_table_view_array_copy.each do |params|
					@transcript_table_view_array << params unless params[:mastercomment2].downcase.index(filter_string).nil? # does contain filter_string?
				end
			when "Person"
				#puts "Person"
				transcript_table_view_array_copy.each do |params|
					@transcript_table_view_array << params unless params[:Person].downcase.index(filter_string).nil? # does contain filter_string?
				end
			when "Event or Theme"
				#puts "Event or Theme"
				transcript_table_view_array_copy.each do |params|
					@transcript_table_view_array << params unless params[:mastercomment4].downcase.index(filter_string).nil? # does contain filter_string?
				end
			end
		end
		@transcriptTableView.reloadData
		@transcriptTableView.deselectAll(self)
	end
	ib_action :filterTranscripts

	def previewTranscript(sender)
		@transcriptSpinner2.startAnimation(self)
		selected_row = @transcriptTableView.selectedRowIndexes.to_a[0]
		play_selected_subclip(@transcript_table_view_array[selected_row]) #unless @transcript_table_view_array[selected_row][:thumbnail]==@offline_image
		@transcriptSpinner2.stopAnimation(self)
	end
	ib_action :previewTranscript

	def joinTranscriptRows(sender) # joins both :Transcript and :timestamps_array in @transcript_table_view_array
		#puts "joinTranscriptRows"
		selected_rows = @transcriptTableView.selectedRowIndexes.to_a
		joinRows( {"tag" => @transcriptTableView.tag, "rows" => selected_rows} )
	end
	ib_action :joinTranscriptRows

	### END TRANSCRIPT VIEW METHODS ###

	### STORY VIEW METHODS ###
	
	def reorderStoryList # AND calculate story_duration
		#puts "reorderStoryList"
		story_duration = 0
		@story_table_view_array.each_with_index do |s, i| 
			s[:story_order] = i+1
			story_duration = story_duration + s[:out_point] - s[:in_point]
		end
		@storyBox.setTitle("Story                                       " + to_time(story_duration))
	end

	def deselectAll(sender)
		@selectsTableView.deselectAll(self)
		@storyTableView.deselectAll(self)
	end
	ib_action :deselectAll
		
	def filterSelects(sender) # called by both the NSSearchField and the popup menu beside it (changing the popup menu does the filter again)
		filter_string = @selectsSearchField.stringValue.to_s.downcase
		#puts "filterSelects: search for #{filter_string.inspect}"
		if filter_string=="" # cancel
			@selects_table_view_array = @subclips_array.find_all { |s| s[:good]=="TRUE" && s[:added_to_story]==false }
		else
			selects_table_view_array_copy = @subclips_array.find_all { |s| s[:good]=="TRUE" && s[:added_to_story]==false }
			@selects_table_view_array = []
			case @searchPopupMenu.titleOfSelectedItem
			when "Everywhere"
				#puts "Everywhere"
				selects_table_view_array_copy.each do |params|
					content_string = params[:Person].downcase + " " + params[:searchable_transcript] + " " + " " + params[:mastercomment1].downcase + params[:mastercomment2].downcase + " " + params[:mastercomment4].downcase
					@selects_table_view_array << params unless content_string.index(filter_string).nil? # does contain filter_string?
				end
			when "Transcript"
				#puts "Transcript"
				selects_table_view_array_copy.each do |params|
					@selects_table_view_array << params unless params[:searchable_transcript].index(filter_string).nil? # does contain filter_string?
				end
			when "Story Keywords"
				#puts "Story Keywords"
				selects_table_view_array_copy.each do |params|
					@selects_table_view_array << params unless params[:mastercomment1].downcase.index(filter_string).nil? # does contain filter_string?
				end
			when "Location"
				#puts "Location"
				selects_table_view_array_copy.each do |params|
					@selects_table_view_array << params unless params[:mastercomment2].downcase.index(filter_string).nil? # does contain filter_string?
				end
			when "Person"
				#puts "Person"
				selects_table_view_array_copy.each do |params|
					@selects_table_view_array << params unless params[:Person].downcase.index(filter_string).nil? # does contain filter_string?
				end
			when "Event or Theme"
				#puts "Event or Theme"
				selects_table_view_array_copy.each do |params|
					@selects_table_view_array << params unless params[:mastercomment4].downcase.index(filter_string).nil? # does contain filter_string?
				end
			end
		end
		@selectsTableView.reloadData
		@selectsTableView.deselectAll(self)
	end
	ib_action :filterSelects

	def previewSelect(sender)
		@selectsSpinner.startAnimation(self)
		selected_row = @selectsTableView.selectedRowIndexes.to_a[0]
		play_selected_subclip(@selects_table_view_array[selected_row]) unless @selects_table_view_array[selected_row][:thumbnail]==@offline_image
		@selectsSpinner.stopAnimation(self)
	end
	ib_action :previewSelect

	def previewStory(sender)
		@storySpinner.startAnimation(self)
		@speechSynth = OSX::NSSpeechSynthesizer.alloc.init
		@speechSynth.setDelegate(self)
		if @storyTableView.numberOfSelectedRows==0
			play_story(@story_table_view_array)
		else
			array_of_subclips = Array.new
			@storyTableView.selectedRowIndexes.to_a.each { |r| array_of_subclips << @story_table_view_array[r] }
			#puts "array_of_subclips.length = #{array_of_subclips.length}"
			play_story(array_of_subclips)
		end
		@storySpinner.stopAnimation(self)
	end
	ib_action :previewStory

	def showThumbnails(sender)
		#puts "showThumbnails"
		if @showThumbnailsCheckbox.state==1
			@storyTableView.tableColumns[0].setDataCell(@imageCell)
		else
			@storyTableView.tableColumns[0].setDataCell(@tokenFieldCell)
		end
		@storyTableView.reloadData
	end
	ib_action :showThumbnails

	def showPaperColors(sender)
		@selectsTableView.reloadData
		@storyTableView.reloadData
	end
	ib_action :showPaperColors

	def addSelectsToStory(sender)
		selected_rows = @selectsTableView.selectedRowIndexes.to_a
		move_subclips_from_selects_to_story(selected_rows, @story_table_view_array.length+1)
	end
	ib_action :addSelectsToStory

	def move_subclips_from_selects_to_story(selected_rows, insert_at_row)
		altered_rows = Array.new # the entire @story_table_view_array will be altered because of the :story_order renumbering
		@story_table_view_array.each { |r| altered_rows << {"clip_id" => r[:clip_id], "subclip_id" => r[:subclip_id], "good" => r[:good].to_s, "added_to_story" => r[:added_to_story].to_s, "story_order" => r[:story_order].to_s} }
		selected_rows.each_with_index do |r, i|
			altered_rows << {"clip_id" => @selects_table_view_array[r][:clip_id], "subclip_id" => @selects_table_view_array[r][:subclip_id], "good" => @selects_table_view_array[r][:good].to_s, "added_to_story" => @selects_table_view_array[r][:added_to_story].to_s, "story_order" => @selects_table_view_array[r][:story_order].to_s} 
			@selects_table_view_array[r][:added_to_story] = true
			@selects_table_view_array[r][:story_order] = insert_at_row + (0.000001 * (i+1)) # move selected_rows fractionally forward in order after the drop row
			@story_table_view_array.insert(0, @selects_table_view_array[r])
		end
		@story_table_view_array = @story_table_view_array.sort_by { |s| s[:story_order] }
		selected_rows.reverse_each { |r| @selects_table_view_array.delete_at(r) }
		@selectsTableView.reloadData
		@selectsTableView.deselectAll(self)
		@storyTableView.reloadData
		@storyTableView.deselectAll(self)
		# reorderStoryList to change the story_order to integers for the list
		reorderStoryList()
		# add to undoManager
		if selected_rows.length==1
			action_name = "Move Subclip"
		else
			action_name = "Move Subclips"
		end
		undoManager.registerUndoWithTarget_selector_object(self, 'changeSubclipsArray', {"undo_array" => altered_rows, "action_name" => action_name} )
		undoManager.setActionName(action_name)
	end

	def removeSelectsFromStory(sender)
		selected_rows = @storyTableView.selectedRowIndexes.to_a
		move_subclips_from_story_to_selects(selected_rows, 0)
	end
	ib_action :removeSelectsFromStory

	def move_subclips_from_story_to_selects(selected_rows, insert_at_row)
		altered_rows = Array.new # the entire @story_table_view_array will be altered because of the :story_order renumbering
		@story_table_view_array.each { |r| altered_rows << {"clip_id" => r[:clip_id], "subclip_id" => r[:subclip_id], "good" => r[:good].to_s, "added_to_story" => r[:added_to_story].to_s, "story_order" => r[:story_order].to_s} }
		selected_rows.each do |r|
			@story_table_view_array[r][:added_to_story] = false
			@story_table_view_array[r][:story_order] = -1
			@selects_table_view_array.insert(insert_at_row, @story_table_view_array[r])
		end
		@selects_table_view_array = @selects_table_view_array.sort { |s1, s2| s1[:clip_id]==s2[:clip_id] ? s1[:subclip_id]<=>s2[:subclip_id] : s1[:clip_id]<=>s2[:clip_id] } # sort by :clip_id then by :subclip_id
		selected_rows.reverse_each { |r| @story_table_view_array.delete_at(r) }
		@selectsTableView.reloadData
		@selectsTableView.deselectAll(self)
		@storyTableView.reloadData
		@storyTableView.deselectAll(self)
		# reorderStoryList to change the story_order to integers for the list
		reorderStoryList()
		# add to undoManager
		if selected_rows.length==1
			action_name = "Move Subclip"
		else
			action_name = "Move Subclips"
		end
		undoManager.registerUndoWithTarget_selector_object(self, 'changeSubclipsArray', {"undo_array" => altered_rows, "action_name" => action_name} )
		undoManager.setActionName(action_name)
	end

	def joinStoryRows(sender)
		#puts "joinStoryRows"
		selected_rows = @storyTableView.selectedRowIndexes.to_a
		joinRows( {"tag" => @storyTableView.tag, "rows" => selected_rows} )
	end
	ib_action :joinStoryRows

	def export(sender)
		@sendToPopup.selectItemAtIndex(@project_array[5])
		frame_rates = Array.new
		@story_table_view_array.each_with_index do |s, i|
			if s[:type].nil?
				this_frame_rate = FRAME_RATES.find { |rate| rate[:ntsc] == s[:ntsc] && rate[:timebase] == s[:timebase] }
				#puts this_frame_rate.inspect
				frame_rates << this_frame_rate[:type]
			else
				frame_rates << s[:type]
			end
		end
		#puts frame_rates.inspect
		# Sequence Settings box
		@frameRatePopup.removeAllItems
		@frameRatePopup.addItemsWithTitles(frame_rates.uniq.sort)
		@frameRatePopup.setEnabled(true)
		slugs = @story_table_view_array.find_all { |s| s[:clip_name]=="Voice Overs" }
		@sendToPopup.itemWithTitle("Save Voice Over Script").setEnabled(true) unless slugs.empty?
		OSX::NSApp.beginSheet_modalForWindow_modalDelegate_didEndSelector_contextInfo( @exportDialog, @documentWindow, self, 'didEndExport:returnCode:contextInfo:', nil )
	end
	ib_action :export

	def closeExportDialog(sender)
		chosen_frame_rate = FRAME_RATES.find { |rate| rate[:type] == @frameRatePopup.titleOfSelectedItem }
		@ntsc = chosen_frame_rate[:ntsc]
		@timebase = chosen_frame_rate[:timebase]
		#puts "closeExportDialog #{@sendToPopup.titleOfSelectedItem}"
		@exportDialog.orderOut(nil)
		OSX::NSApp.endSheet(@exportDialog)
	end
	ib_action :closeExportDialog

	def didEndExport_returnCode_contextInfo(sheet, buttonClicked, info)
		@exportSpinner.startAnimation(self)
		begin
			create_fcp_xml()
			case @sendToPopup.titleOfSelectedItem
			when "Save Excel Spreadsheet"
				if @project_array[5]!=0
					@project_array[5] = 0
					self.updateChangeCount(OSX::NSChangeDone)
				end
				sPanel = OSX::NSSavePanel.savePanel
				sPanel.setRequiredFileType("xls")
				sPanel.setExtensionHidden(false)
				sPanel.setCanSelectHiddenExtension(true)
				sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, "#{@xml_name}", self.windowForSheet, self, 'didEndSaveSpreadsheet:returnCode:contextInfo:', nil)
			when "Save Text Transcript"
				if @project_array[5]!=4
					@project_array[5] = 4
					self.updateChangeCount(OSX::NSChangeDone)
				end
				sPanel = OSX::NSSavePanel.savePanel
				sPanel.setRequiredFileType("txt")
				sPanel.setExtensionHidden(false)
				sPanel.setCanSelectHiddenExtension(true)
				sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, "#{@xml_name} Transcript", self.windowForSheet, self, 'didEndSaveText:returnCode:contextInfo:', nil)
			when "Save Voice Over Script"
				if @project_array[5]!=5
					@project_array[5] = 5
					self.updateChangeCount(OSX::NSChangeDone)
				end
				sPanel = OSX::NSSavePanel.savePanel
				sPanel.setRequiredFileType("txt")
				sPanel.setExtensionHidden(false)
				sPanel.setCanSelectHiddenExtension(true)
				sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, "#{@xml_name} Transcript", self.windowForSheet, self, 'didEndSaveVO:returnCode:contextInfo:', nil)
			when "Send to Final Cut Pro"
				if @project_array[5]!=1
					@project_array[5] = 1
					self.updateChangeCount(OSX::NSChangeDone)
				end
				save_this_file = File.new(TEMP_XML_FILE, "w")
				save_this_file.puts @nokogiri_xml
				save_this_file.close
				OSX::NSWorkspace.sharedWorkspace.openFile_withApplication( TEMP_XML_FILE, "Final Cut Pro" )
				@exportSpinner.stopAnimation(self)
			when "Save XML for Premiere Pro"
				if @project_array[5]!=2
					@project_array[5] = 2
					self.updateChangeCount(OSX::NSChangeDone)
				end
				sPanel = OSX::NSSavePanel.savePanel
				sPanel.setRequiredFileType("xml")
				sPanel.setExtensionHidden(false)
				sPanel.setCanSelectHiddenExtension(false)
				#puts sPanel.inspect
				sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, "#{@xml_name}", @documentWindow, self, 'didEndSaveXML:returnCode:contextInfo:', nil)
			when "Save zipped XML file"
				if @project_array[5]!=3
					@project_array[5] = 3
					self.updateChangeCount(OSX::NSChangeDone)
				end
				sPanel = OSX::NSSavePanel.savePanel
				sPanel.setRequiredFileType("zip")
				sPanel.setExtensionHidden(false)
				sPanel.setCanSelectHiddenExtension(false)
				#puts sPanel.inspect
				sPanel.beginSheetForDirectory_file_modalForWindow_modalDelegate_didEndSelector_contextInfo(nil, "#{@xml_name}", @documentWindow, self, 'didEndZip:returnCode:contextInfo:', nil)
			end
		rescue
			@exportSpinner.stopAnimation(self)
			displayAlert("There was a problem exporting XML.", "Please choose another XML file exported from\nFinal Cut Pro or Premiere Pro.")
			#puts "PROBLEM with create_fcp_xml()"
		end
	end

	def didEndSaveSpreadsheet_returnCode_contextInfo(sPanel, buttonClicked, info)
		if buttonClicked == OSX::NSOKButton
			@exportSpinner.startAnimation(self)
			sPanel.orderOut(self)
			if sPanel.isExtensionHidden?
				extension_hidden = 1
			else
				extension_hidden = 0
			end
			#puts "is extension hidden?", sPanel.isExtensionHidden
			excel_file = sPanel.URL.path.to_s
			begin
				save_to_excel(excel_file)
			rescue
				displayAlert("There was a problem saving an Excel spreadsheet.", "Please choose another XML file exported from\nFinal Cut Pro or Premiere Pro.")
				#puts "RESCUE!!! Problem with save_to_excel()"
			end
		end
		@exportSpinner.stopAnimation(self)
	end

	def didEndSaveText_returnCode_contextInfo(sPanel, buttonClicked, info)
		if buttonClicked == OSX::NSOKButton
			@exportSpinner.startAnimation(self)
			sPanel.orderOut(self)
			if sPanel.isExtensionHidden?
				extension_hidden = 1
			else
				extension_hidden = 0
			end
			#puts "is extension hidden?", sPanel.isExtensionHidden
			text_file = sPanel.URL.path.to_s
			begin
				save_to_text(text_file, false) # vo_filter false - save all
				save_this_file = File.new(text_file, "w")
				save_this_file.puts @text_transcript
				save_this_file.close
			rescue
				displayAlert("There was a problem saving a Text transcript.", "Please choose another XML file exported from\nFinal Cut Pro or Premiere Pro.")
				#puts "RESCUE!!! Problem with save_to_text()"
			end
		end
		@exportSpinner.stopAnimation(self)
	end

	def didEndSaveVO_returnCode_contextInfo(sPanel, buttonClicked, info)
		if buttonClicked == OSX::NSOKButton
			@exportSpinner.startAnimation(self)
			sPanel.orderOut(self)
			if sPanel.isExtensionHidden?
				extension_hidden = 1
			else
				extension_hidden = 0
			end
			#puts "is extension hidden?", sPanel.isExtensionHidden
			text_file = sPanel.URL.path.to_s
			begin
				save_to_text(text_file, true) # vo_filter true
				save_this_file = File.new(text_file, "w")
				save_this_file.puts @text_transcript
				save_this_file.close
			rescue
				displayAlert("There was a problem saving a Text transcript.", "Please choose another XML file exported from\nFinal Cut Pro or Premiere Pro.")
				#puts "RESCUE!!! Problem with save_to_text()"
			end
		end
		@exportSpinner.stopAnimation(self)
	end

	def didEndSaveXML_returnCode_contextInfo(sPanel, buttonClicked, info)
		if buttonClicked == OSX::NSOKButton
			sPanel.orderOut(self)
			xml_file = sPanel.URL.path.to_s
			save_this_file = File.new(xml_file, "w")
			save_this_file.puts @nokogiri_xml
			save_this_file.close
		end
		@exportSpinner.stopAnimation(self)
	end

	def didEndZip_returnCode_contextInfo(sPanel, buttonClicked, info)
		if buttonClicked == OSX::NSOKButton
			sPanel.orderOut(self)
			zip_file = sPanel.URL.path.to_s
			%x{ditto -c -k --sequesterRsrc "#{TEMP_XML_FILE}" "#{zip_file}"}
		end
		@exportSpinner.stopAnimation(self)
	end

	# undo/redo moving of subclips into/out of/within the Story list
	def changeSubclipsArray(params)
		#puts "changeSubclipsArray"
		undo_array, action_name = params["undo_array"].to_ruby, params["action_name"].to_s
		current_state_array = Array.new # records the state of :added_to_story and :story_order for each subclip
		undo_array.each do |altered_row|
			#puts altered_row.inspect # clip_id, subclip_id, added_to_story, story_order
			this_subclip = @subclips_array.find { |s| s[:clip_id] == altered_row["clip_id"] && s[:subclip_id] == altered_row["subclip_id"] }
			current_state_array << {"clip_id" => this_subclip[:clip_id], "subclip_id" => this_subclip[:subclip_id], "good" => this_subclip[:good].to_s, "added_to_story" => this_subclip[:added_to_story].to_s, "story_order" => this_subclip[:story_order].to_s}
			this_subclip[:good] = altered_row["good"]
			this_subclip[:added_to_story] = (altered_row["added_to_story"]=="true") # converts string to boolean
			this_subclip[:story_order] = altered_row["story_order"].to_i
		end
		if action_name=="Rearrange Story"
			@story_table_view_array = @story_table_view_array.sort_by { |s| s[:story_order] }
			@storyTableView.reloadData
			@storyTableView.deselectAll(self)
		else
			@selects_table_view_array, @story_table_view_array, @clips_table_view_array = story_view_data_sources()
			@story_table_view_array.each { |s| puts "#{s[:clip_id]}, #{s[:subclip_id]}, #{s[:added_to_story]}, #{s[:story_order]}" }
			@selectsTableView.reloadData
			@selectsTableView.deselectAll(self)
			@storyTableView.reloadData
			@storyTableView.deselectAll(self)
		end
		# add to undoManager
		#puts current_state_array.inspect
		undoManager.registerUndoWithTarget_selector_object(self, 'changeSubclipsArray', {"undo_array" => current_state_array, "action_name" => action_name} )
		undoManager.setActionName(action_name)
	end
	
	def addSlug(sender) # Add V.O. or Remake V.O.
		#puts "addSlug"
		if @addVOButton.title=="Add V.O."
			selected_row = @storyTableView.selectedRowIndexes.to_a[-1]
			add_slug(selected_row)
			@storyTableView.reloadData
			# reorderStoryList() happens in control_textShouldEndEditing after editing
			@storyTableView.selectRowIndexes_byExtendingSelection( OSX::NSIndexSet.indexSetWithIndex(selected_row + 1), false)
			# add to undoManager
			undoManager.registerUndoWithTarget_selector_object(self, 'removeSlug', {"row" => selected_row+1 } )
			undoManager.setActionName("Add Voice Over")
		else # Remake V.O.
			slugs = @story_table_view_array.find_all { |s| s[:clip_name]=="Voice Overs" }
			@slugs_to_make = slugs.length
			#puts "@slugs_to_make = #{@slugs_to_make}"
			@exportSpinner.startAnimation(self) unless @slugs_to_make==0
			slugs.each do |slug|
				speechSynth = OSX::NSSpeechSynthesizer.alloc.init
				speechSynth.setDelegate(self)
				speechSynth.startSpeakingString_toURL(slug[:Transcript], OSX::NSURL.fileURLWithPath(slug[:pathurl]))
				slug[:thumbnail] = @voiceover_image
			end
		end
	end
	ib_action :addSlug
	
	def removeSlug(params)
		row = params["row"].to_i
		removed_slug = @story_table_view_array.delete_at(row)
		#puts removed_slug.inspect
		@storyTableView.reloadData
		@storyTableView.deselectAll(self)
		reorderStoryList()
		# add to undoManager
		undoManager.registerUndoWithTarget_selector_object(self, 'unRemoveSlug', {"row" => row, "Transcript" => removed_slug[:Transcript], "pathurl" => removed_slug[:pathurl], "timestamps_array" => removed_slug[:timestamps_array], "in_point" => removed_slug[:in_point], "out_point" => removed_slug[:out_point]} )
		undoManager.setActionName("Add Voice Over")
	end
	
	def unRemoveSlug(params)
		row = params["row"].to_i
		add_slug(row-1)
		@story_table_view_array[row][:Transcript] = params["Transcript"].to_s
		@story_table_view_array[row][:searchable_transcript] = params["Transcript"].to_s.downcase
		@story_table_view_array[row][:pathurl] = params["pathurl"].to_s
		@story_table_view_array[row][:timestamps_array] = params["timestamps_array"].to_ary
		@story_table_view_array[row][:in_point] = params["in_point"].to_i
		@story_table_view_array[row][:out_point] = params["out_point"].to_i
		@story_table_view_array[row][:duration] = @story_table_view_array[row][:out_point] - @story_table_view_array[row][:in_point]
		@storyTableView.reloadData
		reorderStoryList()
		# add to undoManager
		undoManager.registerUndoWithTarget_selector_object(self, 'removeSlug', {"row" => row})
		undoManager.setActionName("Add Voice Over")
	end
	
	# NSSpeechSynthesizerDelegate method
	def speechSynthesizer_didFinishSpeaking(sender, success)
		#puts "speechSynthesizer_didFinishSpeaking #{success}"
		if @slugs_to_make > 0
			@slugs_to_make -= 1
			if @slugs_to_make==0
				@exportSpinner.stopAnimation(self)
				@storyTableView.setNeedsDisplay(true) if @showThumbnailsCheckbox.state==1 # redisplays @storyTableView without reloadData
			end
		end
		#puts "@slugs_to_make = #{@slugs_to_make}"
	end
	
	### END STORY VIEW METHODS ###

	### TOOLBAR METHODS ###

	def setupToolbar
		# Create a new toolbar instance, and attach it to our document window.
		@toolbar ||= OSX::NSToolbar.alloc.initWithIdentifier(TOOLBAR_IDENTIFIER)
		# Set up toolbar properties: disallow customization, give a default display mode and don't remember state in user defaults.
		@toolbar.setAllowsUserCustomization(false)
		@toolbar.setAutosavesConfiguration(false)
		@toolbar.setDisplayMode(OSX::NSToolbarDisplayModeIconAndLabel)
		# We are the delegate.
		@toolbar.setDelegate(self)
		# Attach the toolbar to the window.
		@documentWindow.setToolbar(@toolbar)
	end

	def toolbarDefaultItemIdentifiers(toolbar)
		[ TOOLBAR_FIRST_IDENTIFIER, TOOLBAR_SECOND_IDENTIFIER ]
	end

	def toolbarAllowedItemIdentifiers(toolbar)
		[ TOOLBAR_FIRST_IDENTIFIER, TOOLBAR_SECOND_IDENTIFIER ]
	end

	def toolbarSelectableItemIdentifiers(toolbar)
		[ TOOLBAR_FIRST_IDENTIFIER, TOOLBAR_SECOND_IDENTIFIER ]
	end

	def toolbar_itemForItemIdentifier_willBeInsertedIntoToolbar(toolbar, identifier, flag)
		label, tooltip, image, action = 
		case identifier.to_s
		when TOOLBAR_FIRST_IDENTIFIER
			[ "Transcript View", "Transcript View", "TranscriptToolbarImage", 'showTranscriptView:' ]
		when TOOLBAR_SECOND_IDENTIFIER
			[ "Story View", "StoryView", "StoryToolbarImage", 'showStoryView:' ]
		else
			return nil
		end
		toolbarItem = OSX::NSToolbarItem.alloc.initWithItemIdentifier(identifier)
		toolbarItem.setLabel(label)
		toolbarItem.setPaletteLabel(label)
		toolbarItem.setToolTip(tooltip)
		toolbarItem.setImage(OSX::NSImage.imageNamed(image))
		toolbarItem.setTarget(self)
		toolbarItem.setAction(action)
		return toolbarItem
	end

	### END TOOLBAR METHODS ###

	### TABLEVIEW METHODS ### (tag 2 = Transcript, 3 = Selects, 4 = Story)
	
	def numberOfRowsInTableView(tableView)
		#puts "numberOfRowsInTableView"
		case tableView.tag
		when 1
			return @clips_table_view_array.length
		when 2
			@transcript_table_view_array.each_with_index do |params, i|
				height = calculate_height_for_row(params[:Transcript], 1, i)
				params.merge!( { :row_height => height } )
			end
			return @transcript_table_view_array.length
		when 3
			@selects_table_view_array.each_with_index do |params, i|
				height = calculate_height_for_row(params[:Transcript], 1, i)
				params.merge!( { :row_height => height } )
			end
			return @selects_table_view_array.length
		when 4
			@story_table_view_array.each_with_index do |params, i|
				height = calculate_height_for_row(params[:Transcript], 1, i)
				params.merge!( { :row_height => height } )
			end
			if @story_table_view_array.length==0
				@storyPreviewButton.setEnabled(false)
				@exportButton.setEnabled(false)
			else
				@storyPreviewButton.setEnabled(true)
				if @this_rego[0]["active"]
					@exportButton.setEnabled(true)
				else
					@exportButton.setEnabled(false)
				end
			end
			return @story_table_view_array.length
		end
	end
	
	def tableView_objectValueForTableColumn_row(tableView, tableColumn, rowIndex)
		# puts "tableView_objectValueForTableColumn_row"
		case tableView.tag
		when 1
			return @clips_table_view_array[rowIndex]
		when 2
			if @transcript_table_view_array[rowIndex][:good]=="TRUE"
				text_color = @black_color
			else
				text_color = @light_gray_color
			end
			tableColumn.dataCell.setTextColor(text_color)
			tableColumn.dataCell.setBackgroundColor( @bg_colors[@transcript_table_view_array[rowIndex][:bg_color_index]] )
			return @transcript_table_view_array[rowIndex][:Transcript]
		when 3
			if tableColumn.headerCell.stringValue=="Person"
				return @selects_table_view_array[rowIndex][:Person]
			else
				if @showColorsCheckbox.state==1
					tableColumn.dataCell.setBackgroundColor( @bg_colors[@selects_table_view_array[rowIndex][:bg_color_index]] )
				else
					tableColumn.dataCell.setBackgroundColor(@white_color)
				end
				return @selects_table_view_array[rowIndex][:Transcript]
			end
		when 4
			if tableColumn.headerCell.stringValue=="Person"
				if @showThumbnailsCheckbox.state==1
					if @story_table_view_array[rowIndex][:thumbnail].nil?
						return @default_image
					else
						return @story_table_view_array[rowIndex][:thumbnail]
					end
				else
					return @story_table_view_array[rowIndex][:Person]
				end
			else
				if @showColorsCheckbox.state==1
					tableColumn.dataCell.setBackgroundColor( @bg_colors[@story_table_view_array[rowIndex][:bg_color_index]] )
				else
					tableColumn.dataCell.setBackgroundColor(@white_color)
				end
				return @story_table_view_array[rowIndex][:Transcript]
			end
		end
	end
	
	def tableView_setObjectValue_forTableColumn_row(tableView, anObject, tableColumn, rowIndex)
		#puts "tableView_setObjectValue_forTableColumn_row"
		case tableView.tag
		# 1 is not editable
		when 2
			@transcript_table_view_array[rowIndex][:Transcript] = anObject.to_s
		when 3
			if tableColumn.headerCell.stringValue=="Person"
				@selects_table_view_array[rowIndex][:Person] = anObject.to_s
			else
				@selects_table_view_array[rowIndex][:Transcript] = anObject.to_s
			end
		when 4
			if tableColumn.headerCell.stringValue=="Person"
				@story_table_view_array[rowIndex][:Person] = anObject.to_s
			else
				@story_table_view_array[rowIndex][:Transcript] = anObject.to_s
			end
		end
	end
	
	def tableView_heightOfRow(tableView, rowIndex)
		case tableView.tag
		when 1
			return 17.0
		when 2
			if @transcript_table_view_array[rowIndex][:row_height].nil?
				return 17.0
			else
				return @transcript_table_view_array[rowIndex][:row_height]
			end
		when 3
			if @selects_table_view_array[rowIndex][:row_height].nil?
				return 17.0
			else
				return @selects_table_view_array[rowIndex][:row_height]
			end
		when 4
			if @story_table_view_array[rowIndex][:row_height].nil?
				return 17.0
			elsif @story_table_view_array[rowIndex][:row_height] < 64.0 && @showThumbnailsCheckbox.state==1
				return 64.0
			else
				return @story_table_view_array[rowIndex][:row_height]
			end
		end
	end
	
	def tableView_selectionIndexesForProposedSelection(tableView, proposedSelectionIndexes) # invoked before changing selection to allow modification if necessary
		if tableView.tag==1
			#puts "tableView_selectionIndexesForProposedSelection => save_all_fields"
			selected_row = tableView.selectedRow # can be nil
			save_all_fields() unless selected_row.nil?
		end
		return proposedSelectionIndexes
	end
	
	def tableViewSelectionDidChange(notification)
		#puts "tableViewSelectionDidChange"
		# check if selected rows are the same clip and are consecutive
		selected_rows = notification.object.selectedRowIndexes.to_a
		case notification.object.tag
		when 1 # same as changing clip popup menu
			tableView = @clipsTableView
			@project_array[4] = selected_rows[0] # also changes @selected_clip_index
			populate_transcript_view()
		when 2
			tableView = @transcriptTableView
			@colorPopupMenu.menu.itemArray.each do |item|
				item.setState(OSX::NSOffState)
			end
		when 3
			tableView = @selectsTableView
		when 4
			tableView = @storyTableView
		end
		# disable all buttons and selectively enable them
		@transcriptPreviewButton.setEnabled(false)
		@transcriptJoinButton.setEnabled(false)
		@selectsPreviewButton.setEnabled(false)
		@deselectAllButton.setEnabled(false)
		@addButton.setEnabled(false)
		@removeButton.setEnabled(false)
		@storyJoinButton.setEnabled(false)
		@addVOButton.setTitle("Remake V.O.")
		if tableView==@clipsTableView
			enable_lognotes_and_comments(false, [0])
		elsif selected_rows.empty? && tableView==@transcriptTableView
			enable_lognotes_and_comments(false)
		elsif selected_rows.empty?
			set_Good_checkbox_state(1, [])
			@storyPreviewButton.setTitle("Preview Story")
		elsif selected_rows.length==1 && tableView==@transcriptTableView
			@colorPopupMenu.selectItemAtIndex( @transcript_table_view_array[selected_rows[0]][:bg_color_index] )
			@colorPopupMenu.menu.itemAtIndex( @transcript_table_view_array[selected_rows[0]][:bg_color_index] ).setState( OSX::NSOnState ) # need to explicitely do this since we set them all to NSOffState
			@transcriptPreviewButton.setEnabled(true)
			enable_lognotes_and_comments(true, selected_rows)
		elsif selected_rows.length>0 && tableView==@selectsTableView
			@storyTableView.deselectAll(self)
			@deselectAllButton.setEnabled(true)
			@selectsPreviewButton.setEnabled(true)
			@addButton.setEnabled(true)
			set_Good_checkbox_state(2, selected_rows)
		elsif selected_rows.length>0 && tableView==@storyTableView
			@selectsTableView.deselectAll(self)
			@deselectAllButton.setEnabled(true)
			@storyPreviewButton.setTitle("Preview Selected")
			@removeButton.setEnabled(true)
			set_Good_checkbox_state(2, [])
			@addVOButton.setTitle("Add V.O.")
		end
		unless tableView==@selectsTableView || selected_rows.empty? || selected_rows.length==1 # check if the subclips can be joined (they are from the same clip and are consecutive)
			current_row = selected_rows[0]
			@can_join = true
			selected_rows.each do |row|
				unless row==current_row
					if tableView==@transcriptTableView && @transcript_table_view_array[row][:subclip_id] == @transcript_table_view_array[current_row][:subclip_id]+1 # they are consecutive
						current_row = row
					elsif tableView==@storyTableView && @story_table_view_array[row][:clip_id] == @story_table_view_array[current_row][:clip_id] && @story_table_view_array[row][:subclip_id] == @story_table_view_array[current_row][:subclip_id]+1 # they are from the same clip and consecutive
						current_row = row
					else
						@can_join = false
					end
				end
			end
			if tableView==@transcriptTableView
				@colorPopupMenu.selectItemAtIndex(1) # menu divider
				enable_lognotes_and_comments(false, selected_rows) # TODO tell the combo boxes that multiple items are selected
				@transcriptJoinButton.setEnabled(@can_join) # can join if they are consecutive
			elsif tableView==@storyTableView
				@storyJoinButton.setEnabled(@can_join)
			end
			#puts "@can_join = #{@can_join}"
		end
	end
	
	def tableView_writeRowsWithIndexes_toPasteboard(tableView, rowIndexes, pasteboard)
		case tableView.tag
		when 3
			data = OSX::NSKeyedArchiver.archivedDataWithRootObject rowIndexes
			pasteboard.declareTypes_owner [SelectsPasteboard], self
			pasteboard.setData_forType data, SelectsPasteboard
			return true
		when 4
			data = OSX::NSKeyedArchiver.archivedDataWithRootObject rowIndexes
			pasteboard.declareTypes_owner [StoryPasteboard], self
			pasteboard.setData_forType data, StoryPasteboard
			return true
		end
	end
	
	def tableView_validateDrop_proposedRow_proposedDropOperation(tableView, draggingInfo, row, operation) # makes the item always drop between rows
		if operation==OSX::NSTableViewDropAbove
			return OSX::NSDragOperationEvery
		else
			return OSX::NSDragOperationNone
		end
	end
	
	def tableView_acceptDrop_row_dropOperation(tableView, draggingInfo, row, operation)
		#puts "tableView_acceptDrop_row_dropOperation #{tableView.tag.inspect}, #{draggingInfo.inspect}, #{row.inspect}, #{operation.inspect}"
		self.updateChangeCount(OSX::NSChangeDone)
		pasteboard = draggingInfo.draggingPasteboard
		if tableView.tag==4 and pasteboard.types.containsObject? StoryPasteboard # dragging within @storyTableView
			altered_rows = Array.new # the entire @story_table_view_array will be altered because of the :story_order renumbering
			@story_table_view_array.each { |r| altered_rows << {"clip_id" => r[:clip_id], "subclip_id" => r[:subclip_id], "good" => r[:good].to_s, "added_to_story" => r[:added_to_story].to_s, "story_order" => r[:story_order].to_s} }
			# puts altered_rows.inspect
			selected_rows = OSX::NSKeyedUnarchiver.unarchiveObjectWithData( pasteboard.dataForType(StoryPasteboard) ).to_a
			#puts "adding rows #{selected_rows.inspect} to row #{row.inspect}"
			selected_rows.each_with_index { |r, i| @story_table_view_array[r][:story_order] = row + (0.000001 * (i+1)) } # move selected_rows fractionally forward in order after the drop row
			@story_table_view_array = @story_table_view_array.sort_by { |s| s[:story_order] }
			@storyTableView.reloadData
			@storyTableView.deselectAll(self)
			# reorderStoryList to change the story_order to integers for the list
			reorderStoryList()
			# add to undoManager
			undoManager.registerUndoWithTarget_selector_object(self, 'changeSubclipsArray', {"undo_array" => altered_rows, "action_name" => "Rearrange Story"} )
			undoManager.setActionName("Rearrange Story")
			return true
		elsif tableView.tag==4 and pasteboard.types.containsObject? SelectsPasteboard # dragging from @selectsTableView to @storyTableView
			selected_rows = OSX::NSKeyedUnarchiver.unarchiveObjectWithData( pasteboard.dataForType(SelectsPasteboard) ).to_a
			#puts "moving rows #{selected_rows.inspect} from @selectsTableView to row #{row.inspect} of @storyTableView"
			move_subclips_from_selects_to_story(selected_rows, row)
			return true
		elsif tableView.tag==3 # dragging from @storyTableView to @selectsTableView
			selected_rows = OSX::NSKeyedUnarchiver.unarchiveObjectWithData( pasteboard.dataForType(StoryPasteboard) ).to_a
			#puts "moving rows #{selected_rows.inspect} from @storyTableView to row #{row} of @selectsTableView"
			move_subclips_from_story_to_selects(selected_rows, row)
			return true
		else
			return false
		end
	end
	
	def tableView_toolTipForCell_rect_tableColumn_row_mouseLocation(tableView, aCell, rect, tableColumn, row, mouseLocation)
		case tableView.tag
		when 1
			return @clips_table_view_array[row]
		when 2
			return @transcript_table_view_array[row][:Person].gsub(",", ", ") unless @transcript_table_view_array[row][:Person].empty?
		when 3
			return @selects_table_view_array[row][:Person].gsub(",", ", ") unless @selects_table_view_array[row][:Person].empty?
		when 4
			return @story_table_view_array[row][:Person].gsub(",", ", ") unless @story_table_view_array[row][:Person].empty?
		end
	end
	
	### END TABLEVIEW METHODS ###

	### TOKENFIELD METHOD ###
	
	def tokenField_completionsForSubstring_indexOfToken_indexOfSelectedItem( tokenField, substring, tokenIndex, selectedIndex )
		#puts "tokenField_completionsForSubstring_indexOfToken_indexOfSelectedItem"
		case tokenField.tag
		when 11 # mastercomment1
			if substring=="?"
				return @keywords
			else
				matching_keywords = @keywords.find_all { |k| k.index(substring)==0 } # matches at the start
				#puts "matching_keywords #{matching_keywords.inspect}"
				return matching_keywords
			end
		when 13 # mastercomment3
			if substring=="?"
				return @people
			else
				matching_people = @people.find_all { |p| p.index(substring)==0 }
				#puts "matching_people #{matching_people.inspect}"
				return matching_people
			end
		end
	end
	
	### TEXTFIELDCELL METHODS ###
	
	def control_textShouldBeginEditing(control, fieldEditor)
		#puts "\n1 control_textShouldBeginEditing", control.tag
		if control.tag < 10 # NSTableView fieldEditors
			@transcript_before_editing = control.stringValue # control.stringValue is the original value of the fieldEditor's string
		end
		return true
	end
	
	def control_textShouldEndEditing(control, fieldEditor)
		#puts "\n4 control_textShouldEndEditing", control.tag
		if control.tag==11
			#puts "It's the Keywords TokenField"
			addKeywords(self)
			return true
		elsif control.tag==13
			#puts "It's the Person TokenField"
			addPeople(self)
			return true
		elsif @transcript_before_editing.nil?
			return false
		else # NSTableView fieldEditors
			#puts "@transcript_before_editing = #{@transcript_before_editing.inspect}"
			transcript_after_editing = control.stringValue
			case control.tag
			when 2
				subclip = @transcript_table_view_array[@transcriptTableView.editedRow]
			when 4
				subclip = @story_table_view_array[@storyTableView.editedRow]
			else
				subclip = nil
			end
			if subclip!=nil && subclip[:clip_name]=="Voice Overs" # allow editing of Voice Overs text
				current_time = 0
				timestamps_array = Array.new
				transcript_after_editing.to_s.split(" ").each do |word|
					timestamps_array << [current_time, current_time + 334]
					current_time += 334
				end
				subclip[:timestamps_array] = timestamps_array
				subclip[:in_point] = 0
				subclip[:out_point] = current_time
				subclip[:duration] = current_time
				subclip[:Transcript] = transcript_after_editing.to_s.strip
				subclip[:searchable_transcript] = subclip[:Transcript].downcase
				subclip[:row_height] = calculate_height_for_row(subclip[:Transcript], control.tag, @storyTableView.editedRow)
				subclip[:thumbnail] = @default_image
				# make a voiceover AIFF
				#puts "starting Speech Synthesis..."
				speechSynth = OSX::NSSpeechSynthesizer.alloc.init
				# speechSynth.setDelegate(self)
				speechSynth.startSpeakingString_toURL(subclip[:Transcript], OSX::NSURL.fileURLWithPath(subclip[:pathurl]))
				reorderStoryList()
				self.updateChangeCount(OSX::NSChangeDone)
				return true
			elsif @transcript_before_editing.to_s.split(" ").length == transcript_after_editing.to_s.split(" ").length # we've maintained the same number of words
				self.updateChangeCount(OSX::NSChangeDone)
				return true
			else
				fieldEditor.setString(@transcript_before_editing)
				return false
			end
		end
	end
	
	def control_textView_doCommandBySelector(control, fieldEditor, commandSelector)
		#puts "\n control_textView_doCommandBySelector", control.tag, commandSelector.inspect
		case control.tag
		when 2
			@tableView = @transcriptTableView
			@table_view_array = @transcript_table_view_array
		when 3
			@tableView = @selectsTableView
			@table_view_array = @selects_table_view_array
		when 4
			@tableView = @storyTableView
			@table_view_array = @story_table_view_array
		end
		if control.tag==2 || control.tag==4 # editable NSTableViews
			subclip = @table_view_array[@tableView.editedRow]
			case commandSelector
			when "insertNewline:"
				original_string = fieldEditor.string.to_s
				fieldEditor.insertNewlineIgnoringFieldEditor(nil)
				if subclip[:clip_name]=="Voice Overs"
					fieldEditor.setString(original_string)
					return false
				elsif valid_split_position(fieldEditor.string)
					splitRow( {"tag" => @tableView.tag, "row" => @tableView.editedRow, "string" => fieldEditor.string} )
					return true
				else # invalid
					fieldEditor.setString(original_string)
					return false
				end
			when "deleteBackward:" || "deleteForward:"
				location = fieldEditor.selectedRange.location
				length = fieldEditor.selectedRange.length
				selected_string = fieldEditor.string.to_s[location, length]
				letter_before = fieldEditor.string.to_s[location-1, 1]
				letter_after = fieldEditor.string.to_s[location+length, 1]
				#puts letter_before.inspect, letter_after.inspect, selected_string.inspect
				if subclip[:clip_name]=="Voice Overs"
					return false
				elsif selected_string.include?(" ")
					return true # ignore the delete
					# selected_string_array = selected_string.split(" ")
					# deleted_text = selected_string_array.collect! { |s| "~" }.join(" ")
					# fieldEditor.setMarkedText_selectedRange(deleted_text, [0, deleted_text.size]) # Replaces currently marked text in the receiver’s text storage with the given string and sets the selection to the given range, computed from the beginning of the marked text.
					# return true # override the delete
				elsif letter_before==" " || letter_after==" "
					return true # ignore the delete
				else
					return false
				end
			when "cancelOperation:"
				selectedRows = @tableView.selectedRowIndexes
				@tableView.deselectAll(self)
				@tableView.selectRowIndexes_byExtendingSelection(selectedRows, true)
				return true
			end
		else # not an editable NSTableView
			#puts "...use default behavior"
			return false
		end
	end

	### NSOPENPANEL METHOD ###
	
	def panel_shouldShowFilename(sender, filename) 
		# FIXME Deprecated in Mac OS X v10.6. Use panel:shouldEnableURL: (NSOpenSavePanelDelegate, 10.6 onwards). filename = url.path
		if File.directory?(filename)
			return true
		elsif File.basename(filename)==File.basename(@reconnect_pathurl)
			return true
		else
			return false
		end
	end
	
end
